/*
 *
 *  Copyright (C) 1996-2000, OFFIS
 *
 *  This software and supporting documentation were developed by
 *
 *    Kuratorium OFFIS e.V.
 *    Healthcare Information and Communication Systems
 *    Escherweg 2
 *    D-26121 Oldenburg, Germany
 *
 *  THIS SOFTWARE IS MADE AVAILABLE,  AS IS,  AND OFFIS MAKES NO  WARRANTY
 *  REGARDING  THE  SOFTWARE,  ITS  PERFORMANCE,  ITS  MERCHANTABILITY  OR
 *  FITNESS FOR ANY PARTICULAR USE, FREEDOM FROM ANY COMPUTER DISEASES  OR
 *  ITS CONFORMITY TO ANY SPECIFICATION. THE ENTIRE RISK AS TO QUALITY AND
 *  PERFORMANCE OF THE SOFTWARE IS WITH THE USER.
 *
 *  Module:  wlistctn
 *
 *  Author:  Andrew Hewett
 *
 *  Purpose: Worklist Database Test Program
 *
 *  Last Update:      $Author: meichel $
 *  Update Date:      $Date: 2000/03/08 16:37:57 $
 *  Source File:      $Source: /share/dicom/cvs-depot/dcmtk/wlistctn/tests/wltest.cc,v $
 *  CVS/RCS Revision: $Revision: 1.16 $
 *  Status:           $State: Exp $
 *
 *  CVS/RCS Log at end of file
 *
 */

#include "osconfig.h"    /* make sure OS specific configuration is included first */

#ifdef HAVE_STDLIB_H
#ifndef  _BCB4
/* workaround for bug in Borland C++ Builder 4 */
BEGIN_EXTERN_C
#endif
#include <stdlib.h>
#ifndef  _BCB4
END_EXTERN_C
#endif
#endif

BEGIN_EXTERN_C
#include <stdio.h>
#include <string.h>
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <errno.h>
END_EXTERN_C

#include "dicom.h"
#include "wrklstdb.h"
#include "dcfilefo.h"
#include "dcdebug.h"
#include "dcdict.h"
#include "cmdlnarg.h"
#include "dcuid.h"    /* for dcmtk version name */

static char rcsid[] = "$dcmtk: wltest v"
  OFFIS_DCMTK_VERSION " " OFFIS_DCMTK_RELEASEDATE " $";

char* progname = NULL;
OFBool verbose = OFFalse;
int debug = 0;

static void
shortusage()
{
    fprintf(stderr, 
	    "usage: %s [-k key][-v][-d<n>] <worklist-db-dir> <query> [<query> ...]\n", 
	    progname);
}

static void
fullusage()
{
    fprintf(stderr, "%s\n\n", rcsid);
    shortusage();
    fprintf(stderr, "\
parameters:\n\
    <worklist-db-dir>	the worklist database directory\n\
    <query>		a dicom file containing a worklist query\n\
options:\n\
    -k key      override query key (gggg,eeee=\"string\")\n\
    -v		verbose mode\n\
    -d<n>	set debug level <n>\n");
    exit(1);
}

static void 
usage()
{
    shortusage();
    exit(1);
}

void errmsg(const char* msg, ...)
{
    va_list args;
	
    fprintf(stderr, "%s: ", progname);
    va_start(args, msg);
    vfprintf(stderr, msg, args);
    va_end(args);
    fprintf(stderr, "\n");
}


static void
addOverrideKey(DcmDataset& overrideKeys, char* s)
{
    unsigned int g = 0xffff;
    unsigned int e = 0xffff;
    int n = 0;
    char val[1024];

    val[0] = '\0';
    n = sscanf(s, "%x,%x=%s", &g, &e, val);
    
    if (n < 2) {
	errmsg("bad key format: %s", s);
	usage(); /* does not return */
    }

    DcmTag tag(g,e);
    if (tag.error() != EC_Normal) {
	errmsg("unknown tag: (%04x,%04x)", g, e);
	usage();
    }
    DcmElement *elem = newDicomElement(tag);
    if (elem == NULL) {
	errmsg("cannot create element for tag: (%04x,%04x)", g, e);
	usage();
    }
    if (strlen(val) > 0) {
	elem->putString(val);
	if (elem->error() != EC_Normal) {
	    errmsg("cannot put tag value: (%04x,%04x)=\"%s\"", g, e, val);
	    usage();
	}
    }

    overrideKeys.insert(elem, OFTrue);
    if (overrideKeys.error() != EC_Normal) {
	errmsg("cannot insert tag: (%04x,%04x)", g, e);
	usage();
    }
}

static OFBool
fileExists(const char* fname) 
{
    OFBool ok = OFFalse;
    FILE* f = NULL;

    if ((f = fopen(fname, "r")) != NULL) {
	ok = OFTrue;
	fclose(f);
    }
    return ok;
}

static OFBool
queryWorklistDB(WorklistDB& wdb, 
		const char* queryFilename, DcmDataset& overrideKeys);


int main(int argc, char* argv[])
{
    int i = 0;
    int j = 0;

#ifdef HAVE_GUSI_H
    /* needed for Macintosh */
    GUSISetup(GUSIwithSIOUXSockets);
    GUSISetup(GUSIwithInternetSockets);
#endif

    prepareCmdLineArgs(argc, argv, "wltest");

    /* strip any leading path from program name */
    if ((progname = (char *)strrchr(argv[0], PATH_SEPARATOR)) != NULL) {
	progname++;
    } else {
	progname = argv[0];
    }

    if (argc < 2) {
	fullusage();
    }

    /* make sure data dictionary is loaded */
    if (!dcmDataDict.isDictionaryLoaded()) {
	errmsg("Warning: no data dictionary loaded, check environment variable: %s\n",
		DCM_DICT_ENVIRONMENT_VARIABLE);
    }

    /* a set of query keys to be added/substituted in each query */
    DcmDataset overrideKeys;

    /* parse program arguments */
    for (i = 1; i < argc && argv[i][0] == '-'; i++) {
	switch (argv[i][1]) {
	case 'v':
	    verbose = OFTrue;
	    break;
	case 'd':
	    if (((i + 1) < argc) && (argv[i + 1][0] != '-') &&
		(sscanf(argv[i + 1], "%d", &debug) == 1)) {
		i++;		
	    } else {
		debug = 1;
	    }
	    verbose = OFTrue;
	    break;
        case 'k':
	    if ((i + 1) < argc) {
		addOverrideKey(overrideKeys, argv[++i]);
	    } else {
		usage();
	    }
	    break;
	default:
	    fullusage();
	}
    }

#ifdef USE_MALLOC_DEBUG
#ifdef HAVE_MALLOC_DEBUG
#define MALLOCDEBUGLEVEL 16
    if (verbose) {
        printf("malloc_debug on (%d)\n", MALLOCDEBUGLEVEL);
    }
    malloc_debug(MALLOCDEBUGLEVEL);
#endif
#endif

    for (j=i; j<argc; j++) {
	if (!fileExists(argv[j])) {
	    errmsg("cannot access %s: %s", argv[j], 
		   strerror(errno));
	    usage();
	}
    }

    SetDebugLevel((debug));	/* dcmdata debugging */

    char* dbPath = argv[i];
    i++;

    WorklistDB wdb(dbPath);

    for (j=i; j<argc; j++) {
	char* queryFilename = argv[j];

	queryWorklistDB(wdb, queryFilename, overrideKeys);
    }

    return 0;
}

static void
substituteOverrideKeys(DcmDataset* query, DcmDataset& overrideKeys)
{
    if (overrideKeys.card() == 0) {
	return; /* nothing to do */
    }

    /* copy the override keys */
    DcmDataset keys(overrideKeys);

    /* put the override keys into query replacing existing tags */
    int elemCount = (int)(keys.card());
    for (int i=0; i<elemCount; i++) {
	DcmElement *elem = keys.remove((unsigned long)0);

	query->insert(elem, OFTrue);
    }
}

static OFBool
isaPendingStatus(WDBStatus status)
{
    return ((status == WDB_Pending) || (status == WDB_Pending_Warning));
}

static const char*
status2string(WDBStatus status) 
{
    const char* s = NULL;
    switch (status) {
    case WDB_Refused_OutOfResorces:
	s = "Refused_OutOfResorces"; break;
    case WDB_Failed_IdentifierDoesNotMatchSOPClass:
	s = "Failed_IdentifierDoesNotMatchSOPClass"; break;
    case WDB_Failed_UnableToProcess:
	s = "Failed_UnableToProcess"; break;
    case WDB_Cancel:
	s = "Cancel"; break;
    case WDB_Success:
	s = "Success"; break;
    case WDB_Pending:
	s = "Pending"; break;
    case WDB_Pending_Warning:
	s = "Pending_Warning"; break;
    default:
	s = "<UNKNOWN-STATUS>"; break;
    }
    return s;
}

static OFBool
queryWorklistDB(WorklistDB& wdb, 
		const char* queryFilename, DcmDataset& overrideKeys)
{
    
    DcmFileFormat dcmff;

    if (queryFilename != NULL) {
	DcmFileStream inf(queryFilename, DCM_ReadMode);
	if ( inf.Fail() ) {
	    errmsg("Cannot open file: %s: %s", queryFilename, strerror(errno));
	    return OFFalse;
	}

	dcmff.transferInit();
	dcmff.read(inf, EXS_Unknown);
	dcmff.transferEnd();

	if (dcmff.error() != EC_Normal) {
	    errmsg("Bad DICOM file: %s: %s", queryFilename, 
		   dcmErrorConditionToString(dcmff.error()));
	    return OFFalse;
	}
    }

    /* replace specific keys by those in overrideKeys */
    substituteOverrideKeys(dcmff.getDataset(), overrideKeys);

    if (verbose) {
	printf("------------\n");
	printf("Query:\n");
	dcmff.getDataset()->print();
	printf("------------\n");
	printf("Starting Query\n");
    }

    /* perform the query */
    
    WDBStatus status = WDB_Pending;
    int responseCount = 0;

    /*
    ** Start the find request.
    */
    status = wdb.startFindRequest(*dcmff.getDataset());

    while (isaPendingStatus(status)) {
	
	DcmDataset* response=NULL;

	/*
	** Get the next find response if there is one.
	*/
	response = wdb.nextFindResponse(status);
	responseCount++;

	if (isaPendingStatus(status)) {
	    printf("------------\n");
	    printf("Response %d (%s)\n", responseCount, status2string(status));
	    response->print();
	} else if (response->card() > 0) {
	    errmsg("ERROR: non-pending status with a response!");
	}
	delete response;
    }

    if (verbose) {
	printf("------------\n");
	printf("Query Complete (%s)\n", status2string(status));
	if (wdb.getErrorComments() != NULL) {
	    printf("Error Comment:\n");
	    DcmLongString errorComment(*wdb.getErrorComments());
	    errorComment.print();
	}
	if (wdb.getOffendingElements() != NULL) {
	    printf("Offending Elements:\n");
	    DcmAttributeTag offendingElements(*wdb.getOffendingElements());
	    offendingElements.print();
	}
    }

    return OFTrue;
}

/*
** CVS Log
** $Log: wltest.cc,v $
** Revision 1.16  2000/03/08 16:37:57  meichel
** Updated copyright header.
**
** Revision 1.15  2000/02/23 15:14:04  meichel
** Corrected macro for Borland C++ Builder 4 workaround.
**
** Revision 1.14  2000/02/01 11:01:02  meichel
** Avoiding to include <stdlib.h> as extern "C" on Borland C++ Builder 4,
**   workaround for bug in compiler header files.
**
** Revision 1.13  1999/04/26 13:00:14  meichel
** Updated copyright headers in module wlistctn and makefile dependencies.
**
** Revision 1.12  1998/07/24 13:29:14  meichel
** changed WorklistDB::nextFindResponse() parameters
**   in order to avoid the use of an implicit assignment
**   operator for class DcmDataset.
**
** Revision 1.11  1997/09/18 08:11:08  meichel
** Many minor type conflicts (e.g. long passed as int) solved.
**
** Revision 1.10  1997/07/21 09:09:03  andreas
** - Replace all boolean types (BOOLEAN, CTNBOOLEAN, DICOM_BOOL, BOOL)
**   with one unique boolean type OFBool.
**
** Revision 1.9  1997/05/29 17:06:43  meichel
** All dcmtk applications now contain a version string
** which is displayed with the command line options ("usage" message)
** and which can be queried in the binary with the "ident" command.
**
** Revision 1.8  1997/05/22 13:37:31  hewett
** Modified the test for presence of a data dictionary to use the
** method DcmDataDictionary::isDictionaryLoaded().
**
** Revision 1.7  1997/04/18 08:40:45  andreas
** - The put/get-methods for all VRs did not conform to the C++-Standard
**   draft. Some Compilers (e.g. SUN-C++ Compiler, Metroworks
**   CodeWarrier, etc.) create many warnings concerning the hiding of
**   overloaded get methods in all derived classes of DcmElement.
**   So the interface of all value representation classes in the
**   library are changed rapidly, e.g.
**   E_Condition get(Uint16 & value, const unsigned long pos);
**   becomes
**   E_Condition getUint16(Uint16 & value, const unsigned long pos);
**   All (retired) "returntype get(...)" methods are deleted.
**   For more information see dcmdata/include/dcelem.h
**
** Revision 1.6  1997/01/13 15:26:03  hewett
** Corrected usage message.
**
** Revision 1.5  1996/09/27 08:54:35  hewett
** Enclosed system include files with BEGIN_EXTERN_C/END_EXTERN_C
**
** Revision 1.4  1996/09/24 15:50:08  hewett
** Added conditionaly compiled support for the Macintosh
** environment (GUSI library).
**
** Revision 1.3  1996/05/03 10:33:40  hewett
** Added check for existance of file arguments before processing starts.
**
** Revision 1.2  1996/04/29 19:16:23  hewett
** Corrected logic when printing response identifiers.
**
** Revision 1.1.1.1  1996/04/26 14:16:02  hewett
** Initial Import.
**
*/
